/**
 * Copyright (C) 2007-2011 Lawrence Murray
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * @author Lawrence Murray <lawrence@indii.org>
 * $Rev: 308 $
 * $Date: 2011-10-11 21:56:35 +0800 (Tue, 11 Oct 2011) $
 */
#include "TintDocument.hpp"

#include "../../image/FileImageResource.hpp"
#include "../../../gnome/gnome.hpp"

#include "wx/button.h"
#include "wx/sizer.h"
#include "wx/dcmemory.h"
#include "wx/cshelp.h"
#include "wx/config.h"
#include "wx/stdpaths.h"
#include "wx/filename.h"

using namespace indii;

TintDocument::TintDocument(wxWindow* parent, ImageResource* res,
    ClusterModel* model) : wxPanel(parent), res(res), model(model), own(false),
    isSaved(false), havePath(false) {
  model->setForDialog();
  init();
}

TintDocument::TintDocument(wxWindow* parent, const wxString& path) :
    wxPanel(parent) {
  res = new FileImageResource(path);
  model = new ClusterModel(res);
  
  model->prepare();
  model->cluster();
  model->lock();
  
  own = true;
  isSaved = false;
  havePath = false;
  
  init();
  
  model->unlock();
  watch(model);
}

TintDocument::~TintDocument() {
  /* getting assertion error that another event handler has been
   * pushed and not popped at this stage; make sure explicitly that event
   * handler is self */
  SetEventHandler(this);
  if (own) {
    delete model;
    delete res;
  }
}

void TintDocument::notifyClusterChange(const unsigned int i) {
  isSaved = false;
}

void TintDocument::notifyNumClustersChange() {
  isSaved = false;
}

void TintDocument::notifyNumRepetitionsChange() {
  isSaved = false;
}

void TintDocument::notifySaturationThresholdChange() {
  isSaved = false;
}

void TintDocument::notifyMaxPixelsChange() {
  isSaved = false;
}

void TintDocument::notifySaturationDecayChange() {
  isSaved = false;
}

void TintDocument::notifySaturationSoftnessChange() {
  isSaved = false;
}

void TintDocument::notifyGreyscaleChange() {
  isSaved = false;
}

void TintDocument::init() {
  wxBoxSizer* sizer = new wxBoxSizer(wxVERTICAL);
  previewPane = new PreviewPane(this, res, model);
  sizer->Add(previewPane, 1, wxEXPAND);
  this->SetSizer(sizer);
}

void TintDocument::save() {
  /* pre-condition */
  assert (havePath);
  
  unsigned width = res->getWidth();
  unsigned height = res->getHeight();
  
  wxImage img(width, height);
  model->calcFg(width, height, img);
  img.SetOption(wxIMAGE_OPTION_QUALITY, 90);
  wxLogNull logNo;
  isSaved = img.SaveFile(path);
  if (!isSaved) {
    isSaved = img.SaveFile(path, _("image/png"));
    if (!isSaved) {
	    wxMessageBox(_("Error attempting to save file. File not saved."),
  	      _("Error"), wxOK|wxICON_HAND, this);
    } else {
      wxMessageBox(_("Unrecognised file extension. File saved as PNG."),
                   _("Warning"), wxOK|wxICON_INFORMATION, this);
    }
  }
}

void TintDocument::saveAs() {
	wxString savePath;
  wxConfig config(_("tintii"), _("indii.org"));
  if (!config.Read(_("SavePath"), &savePath)) {
		savePath = wxStandardPaths::Get().GetDocumentsDir();
  }

  wxString path = wxFileSelector(wxFileSelectorPromptStr, savePath,
      _("tintiied.jpg"), _("jpg"), _("Image Files ") +
      wxImage::GetImageExtWildcard() + _("|All files (*)|*"),
      wxFD_SAVE|wxFD_OVERWRITE_PROMPT, this);
  if (!path.empty()) {
    wxFileName file(path);
    savePath = file.GetPath();
    config.Write(_("SavePath"), savePath);
    
    setPath(path);
    save();
  }
}

const wxString& TintDocument::getPath() const {
  return this->path;
}

void TintDocument::setPath(const wxString& path) {
  this->path = path;
  havePath = true;
}

BEGIN_EVENT_TABLE(TintDocument, wxPanel)
END_EVENT_TABLE()
