/**
 * Copyright (C) 2007-2011 Lawrence Murray
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * @author Lawrence Murray <lawrence@indii.org>
 * $Rev: 309 $
 * $Date: 2011-10-20 22:15:34 +0800 (Thu, 20 Oct 2011) $
 */
#include "TintWindow.hpp"

#include "TintDocument.hpp"
#include "WindowMenu.hpp"
#include "../../gui/About.hpp"
#include "../../gui/Help.hpp"
#include "../../gui/Registration.hpp"
#include "../../gui/OpenRecentMenu.hpp"
#include "../../reg/Components.hpp"
#include "../../image/FileImageResource.hpp"
#include "../../../gnome/gnome.hpp"

#include "wx/config.h"
#include "wx/button.h"
#include "wx/sizer.h"
#include "wx/textdlg.h"
#include "wx/filedlg.h"
#include "wx/msgdlg.h"
#include "wx/dcmemory.h"
#include "wx/cshelp.h"
#include "wx/filename.h"
#include "wx/stdpaths.h"

using namespace indii;

TintWindow::TintWindow(wxWindow* parent, const Mode mode,
    TintWindowObserver* obs) : wxFrame(parent, wxID_ANY, _("tintii"), wxDefaultPosition, wxDefaultSize,
    (mode == STANDALONE) ? wxCAPTION|wxMINIMIZE_BOX|wxMAXIMIZE_BOX|wxCLOSE_BOX|wxRESIZE_BORDER :
    wxCAPTION|/*wxSTAY_ON_TOP|*/wxRESIZE_BORDER), obs(obs),
    mode(mode), notebook(NULL), doc(NULL) {
  if (mode == STANDALONE) {
    notebook = new wxAuiNotebook(this);
  }
  init(mode);
  able(mode == EMBED);
}

TintWindow::TintWindow(wxWindow* parent, ImageResource* res,
    ClusterModel* model, const Mode mode, TintWindowObserver* obs) :
    wxFrame(parent, wxID_ANY, _("tintii"), wxDefaultPosition, wxDefaultSize,
    (mode == STANDALONE) ? wxCAPTION|wxMINIMIZE_BOX|wxMAXIMIZE_BOX|wxCLOSE_BOX|wxRESIZE_BORDER : wxCAPTION|/*wxSTAY_ON_TOP|*/wxRESIZE_BORDER), obs(obs),
    mode(mode), notebook(NULL), doc(NULL) {
  TintDocument* doc;
  if (res != NULL && model != NULL) {
    wxBeginBusyCursor();
    if (mode == STANDALONE) {
      notebook = new wxAuiNotebook(this);
      doc = new TintDocument(notebook, res, model);
    } else {
      doc = new TintDocument(this, res, model);
      this->doc = doc;
    }
    wxEndBusyCursor();
    init(mode, doc);
  } else {
    init(mode);
  }
  able(mode == EMBED || notebook->GetPageCount() > 0); 
}

TintWindow::TintWindow(wxWindow* parent, const wxString& path,
    const Mode mode, TintWindowObserver* obs) : wxFrame(parent, wxID_ANY, path,
    wxDefaultPosition, wxDefaultSize,
    (mode == STANDALONE) ? wxCAPTION|wxMINIMIZE_BOX|wxMAXIMIZE_BOX|wxCLOSE_BOX|wxRESIZE_BORDER : wxCAPTION|/*wxSTAY_ON_TOP|*/wxRESIZE_BORDER), obs(obs), mode(mode),
    notebook(NULL), doc(NULL) {  
  wxBeginBusyCursor();
  TintDocument* doc;
  if (mode == STANDALONE) {
    notebook = new wxAuiNotebook(this);
    doc = new TintDocument(notebook, path);
  } else {
    doc = new TintDocument(this, path);
    this->doc = doc;
  }
  wxEndBusyCursor();

  init(mode, doc);
  able(mode == EMBED || notebook->GetPageCount() > 0);
}

TintWindow::~TintWindow() {
  /* save AUI perspective */
  wxConfig config(_("tintii"), _("indii.org"));
  wxString persp = manager->SavePerspective();
  #ifndef WX_28
  config.Write(wxString("Perspective"), persp);
  #else
  config.Write(wxString("Perspective", wxConvUTF8), persp);
  #endif

  manager->UnInit();
}

void TintWindow::layoutDefault() {
  wxAuiPaneInfo& thumbPaneInfo = manager->GetPane(thumbPane);
  thumbPane->Fit();
  thumbPaneInfo.FloatingSize(thumbPane->GetMinSize().GetWidth(), thumbPane->GetMinSize().GetHeight() + EXTRA_PANE_HEIGHT);
  thumbPaneInfo.Right();
  thumbPaneInfo.Layer(2);
  thumbPaneInfo.Dock();
  thumbPaneInfo.Show();
  
  wxAuiPaneInfo& prePaneInfo = manager->GetPane(prePane);
  prePane->Fit();
  prePaneInfo.FloatingSize(prePane->GetMinSize().GetWidth(), prePane->GetMinSize().GetHeight() + EXTRA_PANE_HEIGHT);
  prePaneInfo.Bottom();
  prePaneInfo.Position(2);
  prePaneInfo.Layer(1);
  prePaneInfo.Row(1);
  prePaneInfo.Dock();
  prePaneInfo.Show(false);

  wxAuiPaneInfo& postPaneInfo = manager->GetPane(postPane);
  postPane->Fit();
  postPaneInfo.FloatingSize(postPane->GetMinSize().GetWidth(), postPane->GetMinSize().GetHeight() + EXTRA_PANE_HEIGHT);
  postPaneInfo.Bottom();
  postPaneInfo.Position(1);
  postPaneInfo.Layer(1);
  postPaneInfo.Row(1);
  postPaneInfo.Dock();
  postPaneInfo.Show();

  wxAuiPaneInfo& mixerPaneInfo = manager->GetPane(mixerPane);
  mixerPane->Fit();
  mixerPaneInfo.FloatingSize(mixerPane->GetMinSize().GetWidth(), mixerPane->GetMinSize().GetHeight() + EXTRA_PANE_HEIGHT);
  mixerPaneInfo.Bottom();
  mixerPaneInfo.Position(0);
  mixerPaneInfo.Layer(1);
  mixerPaneInfo.Row(1);
  mixerPaneInfo.Dock();
  mixerPaneInfo.Show();

  manager->Update();
}

int TintWindow::GetReturnCode() {
  return returnCode;
}

void TintWindow::SetReturnCode(const int returnCode) {
  this->returnCode = returnCode;
}

void TintWindow::OnSystemClose(wxCloseEvent& evt) {
  int ret, unsaved = 0;
  bool vetoed = false;
  size_t i;
  TintDocument* doc;

  if (mode == STANDALONE) {
    /* count number of unsaved images */
    for (i = 0; i < notebook->GetPageCount(); ++i) {
      doc = dynamic_cast<TintDocument*>(notebook->GetPage(i));
      if (!doc->saved()) {
        ++unsaved;
      }
    }
    
    /* prompt about unsaved changes as appropriate */
    if (unsaved > 1) {
      #ifdef __WXOSX__
      ret = wxMessageBox(_("If you don't review your images, all your "
          "changes will be lost."), wxString(_("You have ")) << unsaved <<
          wxString(_(" images with unsaved changes. Do you want to review "
          "these changes before quitting?")),
          wxYES_NO|wxCANCEL|wxICON_QUESTION, this);
      #else
      ret = wxMessageBox(wxString(_("You have ")) << unsaved <<
          _(" images with unsaved changes. Do you want to review "
          "these changes before quitting?\n\nIf you don't review your images, "
          "all your changes will be lost."), _("Save changes?"),
          wxYES_NO|wxCANCEL|wxICON_QUESTION, this);
      #endif
    } else {
      ret = wxYES;
    }
    
    if (ret == wxYES) {
      /* check for save */
      for (i = 0; i < notebook->GetPageCount() && !vetoed; ++i) {
        doc = dynamic_cast<TintDocument*>(notebook->GetPage(i));
        if (!doc->saved() && evt.CanVeto()) {
          notebook->SetSelection(i);
          ret = wxMessageBox(_("Save changes to image before closing?"),
              _("Save changes?"), wxYES_NO|wxCANCEL|wxICON_QUESTION, this);
          if (ret == wxYES) {

            if (!doc->hasPath()) {
            	doc->saveAs();
            } else {
              doc->save();
            }
            if (!doc->saved()) {
              evt.Veto();
              vetoed = true;
            }
          } else if (ret == wxCANCEL) {
            evt.Veto();
            vetoed = true;
	        }
        }
      }
    } else if (ret == wxCANCEL) {
      evt.Veto();
      vetoed = true;
    }
  }
  
  if (!vetoed) {
    if (obs != NULL) {
      obs->notifyClose(this);
    }
    Destroy();
  }
}

void TintWindow::OnAllowNotebookDnD(wxAuiNotebookEvent& evt) {
  evt.Allow();
}

void TintWindow::OnMenuOpen(wxMenuEvent& evt) {
  if (evt.GetMenu() == mnFile) {
    mnRecent->OnShow(evt);
  } else if (evt.GetMenu() == mnWindow) {
    mnWindow->OnShow(evt);
  } else if (evt.GetMenu() == mnPopupWindow) {
    mnPopupWindow->OnShow(evt);
  }
}

void TintWindow::OnButtonOpen(wxCommandEvent& evt) {
  wxString openPath;
  {
    wxConfig config(_("tintii"), _("indii.org"));
    if (!config.Read(_("OpenPath"), &openPath)) {
      openPath = wxStandardPaths::Get().GetDocumentsDir();
    }
  }
      
  wxString path = wxFileSelector(wxFileSelectorPromptStr, openPath, wxEmptyString,
      wxEmptyString, _("Image Files ") + wxImage::GetImageExtWildcard() +
      _("|All files (*)|*"), wxFD_OPEN|wxFD_CHANGE_DIR|wxFD_FILE_MUST_EXIST|
      wxFD_PREVIEW, this);
  if (!path.empty()) {
    if (wxImage::CanRead(path)) {
	    wxFileName filename(path);
  	  if (filename.IsOk() && filename.FileExists()) {      
	  	  wxBeginBusyCursor();
  	 	  TintDocument* doc = new TintDocument(notebook, path);
    		wxEndBusyCursor();

    		notebook->AddPage(doc, filename.GetFullName(), true);
    		doc->previewPane->getPreviewImage()->zoomFit();
      
     	  openPath = filename.GetPath();
        {
          wxConfig config(_("tintii"), _("indii.org"));
     	    config.Write(_("OpenPath"), openPath);
        }
        
        OpenRecentMenu::AddFile(path);
      } else {
        wxMessageBox(_("The selected file does not exist. Please try another."),
            _("File does not exist"), wxOK|wxICON_ERROR, this);
      }
    } else {
      wxMessageBox(_("The selected file is of an unsupported type. Please try another."),
                   _("Unsupported type"), wxOK|wxICON_ERROR, NULL);
    }
  }
  able(notebook->GetPageCount() > 0);
}
  
void TintWindow::OnOpenRecent(wxCommandEvent& evt) {
	int id = evt.GetId();
  if (id >= 300 && id < 310) {
		wxString path(OpenRecentMenu::GetFile(id));
    if (wxImage::CanRead(path)) {
  		wxFileName filename(path);
    	if (!path.empty() && filename.IsOk() && filename.FileExists()) {
      	wxBeginBusyCursor();
   	  	TintDocument* doc = new TintDocument(notebook, path);
   	  	wxEndBusyCursor();
    
      	notebook->AddPage(doc, filename.GetFullName(), true);
      	doc->previewPane->getPreviewImage()->zoomFit();
    
      	OpenRecentMenu::AddFile(path);
    	} else {
      	wxMessageBox(_("The selected file does not exist. Please try another."),
        	  _("File does not exist"), wxOK|wxICON_ERROR, this);
      	OpenRecentMenu::RemoveFile(path);
    	}
    } else {
      wxMessageBox(_("The selected file is of an unsupported type. Please try another."),
                   _("Unsupported type"), wxOK|wxICON_ERROR, NULL);
      OpenRecentMenu::RemoveFile(path);
    }
  }
  able(notebook->GetPageCount() > 0);
}

void TintWindow::OnClearRecent(wxCommandEvent& evt) {
  mnRecent->OnClearRecent(evt);
}

void TintWindow::OnButtonClose(wxCommandEvent& evt) {
  int ret;
  bool vetoed = false;
  TintDocument* doc = dynamic_cast<TintDocument*>(
      notebook->GetPage(notebook->GetSelection()));
  if (!doc->saved()) {
    ret = wxMessageBox(_("Save changes to image before closing?"),
        _("Save changes?"), wxYES_NO|wxCANCEL|wxICON_QUESTION, this);
    if (ret == wxYES) {
      if (!doc->hasPath()) {
        doc->saveAs();
      } else {
        doc->save();
      }
      if (!doc->saved()) {
        vetoed = true;
      }
    } else if (ret == wxCANCEL) {
      vetoed = true;
    }
  }
  if (!vetoed) {
    notebook->DeletePage(notebook->GetSelection());
    able(notebook->GetPageCount() > 0);
  }
}

void TintWindow::OnButtonSave(wxCommandEvent& evt) {
  TintDocument* doc = dynamic_cast<TintDocument*>(
      notebook->GetPage(notebook->GetSelection()));
  if (!doc->hasPath()) {
    doc->saveAs();
    if (doc->saved()) {
    	wxFileName filename(doc->getPath());
      notebook->SetPageText(notebook->GetSelection(), filename.GetFullName());
    }
  } else {
    doc->save();
  }
}

void TintWindow::OnButtonSaveAs(wxCommandEvent& evt) {
  TintDocument* doc = dynamic_cast<TintDocument*>(
      notebook->GetPage(notebook->GetSelection()));
  doc->saveAs();
  if (doc->saved()) {
    wxFileName filename(doc->getPath());
    notebook->SetPageText(notebook->GetSelection(), filename.GetFullName());
  }
}

void TintWindow::OnButtonExit(wxCommandEvent& evt) {
  Close();
}

void TintWindow::OnButtonOk(wxCommandEvent& evt) {
  SetReturnCode(wxID_OK);
  Close();
}

void TintWindow::OnButtonCancel(wxCommandEvent& evt) {
  SetReturnCode(wxID_CANCEL);
  Close();
}

void TintWindow::OnButtonAbout(wxCommandEvent& evt) {
  showAboutDialog(this);
}

void TintWindow::OnButtonDefault(wxCommandEvent& evt) {
  TintDocument* doc;
  if (mode == STANDALONE) {
    doc = dynamic_cast<TintDocument*>(notebook->GetPage(notebook->GetSelection()));
  } else {
    doc = this->doc;
  }
  doc->model->setDefaults();
}

void TintWindow::OnButZoomOut(wxCommandEvent& evt) {
  TintDocument* doc;
  if (mode == STANDALONE) {
    doc = dynamic_cast<TintDocument*>(notebook->GetPage(notebook->GetSelection()));
  } else {
    doc = this->doc;
  }
  doc->previewPane->getPreviewImage()->zoomOut();
}

void TintWindow::OnButZoomIn(wxCommandEvent& evt) {
  TintDocument* doc;
  if (mode == STANDALONE) {
    doc = dynamic_cast<TintDocument*>(notebook->GetPage(notebook->GetSelection()));
  } else {
    doc = this->doc;
  }
  doc->previewPane->getPreviewImage()->zoomIn();
}

void TintWindow::OnButZoomNormal(wxCommandEvent& evt) {
  TintDocument* doc;
  if (mode == STANDALONE) {
    doc = dynamic_cast<TintDocument*>(notebook->GetPage(notebook->GetSelection()));
  } else {
    doc = this->doc;
  }
  doc->previewPane->getPreviewImage()->zoomNormal();
}

void TintWindow::OnButZoomFit(wxCommandEvent& evt) {
  TintDocument* doc;
  if (mode == STANDALONE) {
    doc = dynamic_cast<TintDocument*>(notebook->GetPage(notebook->GetSelection()));
  } else {
    doc = this->doc;
  }
  doc->previewPane->getPreviewImage()->zoomFit();
}

void TintWindow::OnButtonRegister(wxCommandEvent& evt) {
  TintRegistration* dlgRegister = new TintRegistration(this, (mode == EMBED) ? Components::PLUGIN : Components::APP);
  dlgRegister->Centre();
  dlgRegister->ShowModal();
  dlgRegister->Destroy();
}

void TintWindow::OnButtonWindow(wxCommandEvent& evt) {
  #ifdef __WXMSW__
  mnPopupWindow->OnShow(wxMenuEvent());
  #endif
  this->PopupMenu(mnPopupWindow);
}

void TintWindow::OnButtonHelp(wxCommandEvent& evt) {
  Help::show();
}

void TintWindow::OnPageClose(wxAuiNotebookEvent& evt) {
  int ret;
  bool vetoed = false;
  TintDocument* doc = dynamic_cast<TintDocument*>(
      notebook->GetPage(notebook->GetSelection()));
  if (!doc->saved()) {
    ret = wxMessageBox(_("Save changes to image before closing?"),
        _("Save changes?"), wxYES_NO|wxCANCEL|wxICON_QUESTION, this);
    if (ret == wxYES) {
      if (!doc->hasPath()) {
        doc->saveAs();
      } else {
        doc->save();
      }
      if (!doc->saved()) {
        evt.Veto();
        vetoed = true;
      }
    } else if (ret == wxCANCEL) {
      evt.Veto();
      vetoed = true;
    }
  }
  if (!vetoed) {
    able(notebook->GetPageCount() > 1);
  }
}

void TintWindow::OnPageChanged(wxAuiNotebookEvent& evt) {
  TintDocument* doc;
  doc = dynamic_cast<TintDocument*>(notebook->GetPage(notebook->GetSelection()));
  doc->previewPane->getPreviewImage()->refresh();
  thumbPane->setModel(doc->model);
  prePane->setModel(doc->model);
  postPane->setModel(doc->model);
  mixerPane->setModel(doc->model);
}

void TintWindow::OnPreProcessPane(wxCommandEvent& evt) {
  if (mode == EMBED) {
    mnPopupWindow->OnPreProcessPane(evt);
  } else {
    mnWindow->OnPreProcessPane(evt);
  }
}

void TintWindow::OnPostProcessPane(wxCommandEvent& evt) {
  if (mode == EMBED) {
    mnPopupWindow->OnPostProcessPane(evt);
  } else {
    mnWindow->OnPostProcessPane(evt);
  }
}

void TintWindow::OnMixerPane(wxCommandEvent& evt) {
  if (mode == EMBED) {
    mnPopupWindow->OnMixerPane(evt);
  } else {
    mnWindow->OnMixerPane(evt);
  }
}

void TintWindow::OnThumbsPane(wxCommandEvent& evt) {
  if (mode == EMBED) {
    mnPopupWindow->OnThumbsPane(evt);
  } else {
    mnWindow->OnThumbsPane(evt);
  }
}

void TintWindow::OnRestoreDefaults(wxCommandEvent& evt) {
  if (mode == EMBED) {
    mnPopupWindow->OnRestoreDefaults(evt);
  } else {
    mnWindow->OnRestoreDefaults(evt);
  }
}

void TintWindow::init(const Mode mode, TintDocument* doc) {    
  if (mode == STANDALONE) {
    /* menu bar */
    wxMenuBar* mbMenu = new wxMenuBar();

    mnFile = new wxMenu();
    mnRecent = new OpenRecentMenu();
    #ifdef __WXOSX__
    /* cloverleaf seems missing in default shortcuts, so explicitly encode */
    mnFile->Append(wxID_OPEN, _("&Open...\tCtrl-O"));
    mnFile->AppendSubMenu(mnRecent, _("Open Recent"));
    mnFile->AppendSeparator();
    mnFile->Append(wxID_CLOSE, _("&Close\tCtrl-W"));
    mnFile->Append(wxID_SAVE, _("&Save\tCtrl-S"));
    mnFile->Append(wxID_SAVEAS, _("Save &As...\tShift-Ctrl-S"));
    mnFile->Append(wxID_EXIT); // moved to program menu
    #else
    mnFile->Append(wxID_OPEN);
    mnFile->AppendSubMenu(mnRecent, _("Open Recent"));
    mnFile->AppendSeparator();
    mnFile->Append(wxID_CLOSE);
    mnFile->Append(wxID_SAVE);
    mnFile->Append(wxID_SAVEAS);
    mnFile->AppendSeparator();
    mnFile->Append(wxID_EXIT);
    #endif
    mbMenu->Append(mnFile, _("&File"));
  
	  wxMenu* mnEdit = new wxMenu();
 	  mnEdit->Append(ID_BUT_DEFAULT, _("&Restore Defaults"));
    
    wxMenu* mnView = new wxMenu();
    #ifdef __WXOSX__
    mnView->Append(wxID_ZOOM_IN, _("Zoom &In\tCtrl-+"));
  	mnView->Append(wxID_ZOOM_OUT, _("Zoom &Out\tCtrl--"));
  	mnView->Append(wxID_ZOOM_100, _("&Normal"));
  	mnView->Append(wxID_ZOOM_FIT, _("&Fit"));
  	#else
  	mnView->Append(wxID_ZOOM_IN);
  	mnView->Append(wxID_ZOOM_OUT);
  	mnView->Append(wxID_ZOOM_100);
  	mnView->Append(wxID_ZOOM_FIT);
  	#endif

  	mnWindow = new WindowMenu(this);

  	wxMenu* mnHelp = new wxMenu();
  	#ifdef __WXOSX__
  	mnHelp->Append(wxID_HELP, _T("&Help\tCtrl-?"));
  	#else
  	mnHelp->Append(wxID_HELP);
  	#endif
  	mnHelp->Append(wxID_ABOUT);
  	if ((mode == EMBED && !Components::isActivated(Components::PLUGIN)) ||
        (mode == STANDALONE && !Components::isActivated(Components::APP))) {
   	 mnHelp->Append(ID_BUT_REGISTER, _("Register"));
  	}

  	mbMenu->Append(mnEdit, _("&Edit"));
  	mbMenu->Append(mnView, _("&View"));
  	mbMenu->Append(mnWindow, _("&Window"));
  	mbMenu->Append(mnHelp, _("&Help"));
  
  	this->SetMenuBar(mbMenu);
  }
  mnPopupWindow = new WindowMenu(this);
  
  /* toolbar */
  wxToolBar* tbTop = this->CreateToolBar(wxNO_BORDER|/*wxTB_TEXT|*/wxTB_HORIZONTAL);
  tbTop->SetToolBitmapSize(wxSize(32,32));
  if (mode == STANDALONE) {
    tbTop->AddTool(wxID_OPEN, _("Open"), *_img_gnome_document_open, _("Open"));
    tbTop->AddTool(wxID_SAVE, _("Save"), *_img_gnome_document_save, _("Save"));
    tbTop->AddTool(wxID_SAVEAS, _("Save As"), *_img_gnome_document_save_as, _("Save As"));
    tbTop->AddSeparator();
  }
  tbTop->AddTool(wxID_ZOOM_IN, _("Zoom In"), *_img_gnome_zoom_in, _("Zoom In"));
  tbTop->AddTool(wxID_ZOOM_OUT, _("Zoom Out"), *_img_gnome_zoom_out, _("Zoom Out"));
  tbTop->AddTool(wxID_ZOOM_100, _("Actual Size"), *_img_gnome_zoom_original, _("Actual Size"));
  tbTop->AddTool(wxID_ZOOM_FIT, _("Zoom to Fit"), *_img_gnome_zoom_best_fit, _("Zoom to Fit"));
  tbTop->AddSeparator();
  tbTop->AddTool(ID_BUT_DEFAULT, _("Restore Defaults"), *_img_gnome_system_run, _("Restore Defaults"));
  if (mode == EMBED) {
	  tbTop->AddSeparator();
 	  tbTop->AddTool(ID_BUT_WINDOW, _("Windows"), *_img_gnome_preferences_system_windows, _("Windows"));
  }
  tbTop->AddSeparator();

  #ifdef __WXOSX__
	/* help system causing crash under Photoshop on Mac, to do with menu perhaps? */
  if (mode != EMBED) {
	  tbTop->AddTool(wxID_HELP, _("Help"), *_img_gnome_help_browser, _("Help"));
  }
  #else
  tbTop->AddTool(wxID_HELP, _("Help"), *_img_gnome_help_browser, _("Help"));
  #endif
  if (mode == EMBED) {
    tbTop->AddTool(wxID_ABOUT, _("About"), *_img_gnome_dialog_information, _("About"));
  }
  if (mode == STANDALONE && !Components::isActivated(Components::APP)) {
    tbTop->AddTool(ID_BUT_REGISTER, _("Register"), *_img_gnome_emblem_web, _("Register"));
  }
  #ifdef __WXOSX__
  if (mode != EMBED) {
    #ifndef WX_28
    tbTop->AddStretchableSpace();
    tbTop->AddControl(new wxContextHelpButton(tbTop), _T("Context help"));
    #else
    tbTop->AddSeparator();
    tbTop->AddControl(new wxContextHelpButton(tbTop));
    #endif
  }
  #else
  #ifndef WX_28
  tbTop->AddStretchableSpace();
  tbTop->AddControl(new wxContextHelpButton(tbTop), _T("Context help"));
  #else
  tbTop->AddSeparator();
  tbTop->AddControl(new wxContextHelpButton(tbTop));//, _T("Context help"));
  #endif
  #endif
  tbTop->Realize();  
  
  /* AUI manager */
  manager = new wxAuiManager(this);
  
  int width, height;
  wxAuiPaneInfo paneInfo;
  paneInfo.DefaultPane();
  paneInfo.CloseButton(true);
  //paneInfo.PinButton(true);
 
  /* notebook */
  paneInfo.CaptionVisible(false);
  paneInfo.Centre();
  paneInfo.PaneBorder(false);

  if (mode == STANDALONE) {
    manager->AddPane(notebook, paneInfo);
    // doc added later, after other panes...
  } else {
    assert (doc != NULL);
    manager->AddPane(doc, paneInfo);
  }
  
  paneInfo.PaneBorder(true);
  paneInfo.CaptionVisible(true);
 
  /* thumbnail pane */
  thumbPane = new ThumbPane(this);
  width = MIN_THUMB_PANE_WIDTH;
  height = 512;
  
  #ifndef WX_28
  paneInfo.Name(_("Thumbs"));
  #else
  paneInfo.Name(wxString("Thumbs", wxConvUTF8));
  #endif
  paneInfo.MinSize(wxSize(width, 110));
  paneInfo.BestSize(wxSize(width, height));
  paneInfo.FloatingSize(width + EXTRA_PANE_WIDTH, height + EXTRA_PANE_HEIGHT);
  paneInfo.Caption(_("Thumbs"));
  paneInfo.Right();
  paneInfo.Layer(2);
  manager->AddPane(thumbPane, paneInfo);
  
  /* preprocessing pane */
  prePane = new PreProcessPane(this);
  width = prePane->GetMinSize().GetWidth();
  height = prePane->GetMinSize().GetHeight();
  
  #ifndef WX_28
  paneInfo.Name(_("Preprocessing"));
  #else
  paneInfo.Name(wxString("Preprocessing", wxConvUTF8));
  #endif
  paneInfo.MinSize(wxSize(width, height));
  paneInfo.BestSize(wxSize(width, height));
  paneInfo.FloatingSize(width + EXTRA_PANE_WIDTH, height + EXTRA_PANE_HEIGHT);
  paneInfo.Caption(_("Preprocessing"));
  paneInfo.Bottom();
  paneInfo.Position(2);
  paneInfo.Layer(1);
  paneInfo.Show(false);
  manager->AddPane(prePane, paneInfo);
  paneInfo.Show(true);
  
  /* postprocessing pane */
  postPane = new PostProcessPane(this);
  width = postPane->GetMinSize().GetWidth();
  height = postPane->GetMinSize().GetHeight();
  
  #ifndef WX_28
  paneInfo.Name(_("Postprocessing"));
  #else
  paneInfo.Name(wxString("Postprocessing", wxConvUTF8));
  #endif
  paneInfo.MinSize(wxSize(width, height));
  paneInfo.BestSize(wxSize(width, height));
  paneInfo.FloatingSize(width + EXTRA_PANE_WIDTH, height + EXTRA_PANE_HEIGHT);
  paneInfo.Caption(_("Postprocessing"));
  paneInfo.Bottom();
  paneInfo.Position(1);
  paneInfo.Layer(1);
  manager->AddPane(postPane, paneInfo);
  
  /* channel mixer pane */
  mixerPane = new ChannelMixerPane(this);
  width = mixerPane->GetMinSize().GetWidth();
  height = mixerPane->GetMinSize().GetHeight();
  
  #ifndef WX_28
  paneInfo.Name(_("Mixer"));
  #else
  paneInfo.Name(wxString("Mixer", wxConvUTF8));
  #endif
  paneInfo.MinSize(wxSize(width, height));
  paneInfo.BestSize(wxSize(width, height));
  paneInfo.FloatingSize(width + EXTRA_PANE_WIDTH, height + EXTRA_PANE_HEIGHT);
  paneInfo.Caption(_("Channel mixer"));
  paneInfo.Bottom();
  paneInfo.Position(0);
  paneInfo.Layer(1);
  manager->AddPane(mixerPane, paneInfo);

	/* add first document to notebook, if given */
  if (mode == STANDALONE && doc != NULL) {
    notebook->AddPage(doc, _T("Unsaved"));
  }
  
  /* load AUI perspective */
  wxConfig config(_("tintii"), _("indii.org"));
  wxString persp;
  if (config.Read(_("Perspective"), &persp)) {
    manager->LoadPerspective(persp);
  }
  
  /* main view seems to disappear occasionally (observed on both wxGTK and
   * wxOSX), so make sure it's shown */
  if (mode == STANDALONE) {
    manager->GetPane(notebook).Show();
  } else {
    manager->GetPane(doc).Show();
  }
    
  /* dialog buttons -- loaded after restoring perspective to ensure they're
   * always placed appropriately */
  if (mode == EMBED) {
    wxPanel* buttonPanel = new wxPanel(this);
    wxStdDialogButtonSizer* buttonSizer = new wxStdDialogButtonSizer();
    
    wxButton* butOk = new wxButton(buttonPanel, wxID_OK);
    buttonSizer->AddButton(butOk);
    wxButton* butCancel = new wxButton(buttonPanel, wxID_CANCEL);
    buttonSizer->AddButton(butCancel);
    #ifndef __WXOSX__
    wxButton* butHelp = new wxButton(buttonPanel, wxID_HELP);
    buttonSizer->AddButton(butHelp);
    #endif
    buttonSizer->Realize();
    
    buttonPanel->SetSizerAndFit(buttonSizer);
    width = buttonPanel->GetMinSize().GetWidth();
    height = buttonPanel->GetMinSize().GetHeight();

    wxAuiPaneInfo paneInfo;
    paneInfo.DefaultPane();
    #ifndef WX_28
    paneInfo.Name(_("Buttons"));
    #else
    paneInfo.Name(wxString("Buttons", wxConvUTF8));
    #endif
    paneInfo.MinSize(wxSize(width, height));
    paneInfo.BestSize(wxSize(width, height));
    paneInfo.CaptionVisible(false);
    paneInfo.Resizable(false);
    paneInfo.Bottom();
    paneInfo.PaneBorder(false);
    paneInfo.Position(0);
    paneInfo.Layer(10);
    
    manager->AddPane(buttonPanel, paneInfo);
  }

  manager->Update();
  SetSize(1000,640);
  SetReturnCode(wxID_CANCEL);

  if (doc != NULL) {
    doc->previewPane->getPreviewImage()->zoomFit();
    thumbPane->setModel(doc->model);
    prePane->setModel(doc->model);
    postPane->setModel(doc->model);
    mixerPane->setModel(doc->model);    
  }
}

void TintWindow::able(const bool enable) {
  wxToolBar* tbTop = this->GetToolBar();
  if (mode == STANDALONE) {
    tbTop->EnableTool(wxID_SAVE, enable);
    tbTop->EnableTool(wxID_SAVEAS, enable);
  }
  tbTop->EnableTool(wxID_ZOOM_OUT, enable);
  tbTop->EnableTool(wxID_ZOOM_IN, enable);
  tbTop->EnableTool(wxID_ZOOM_100, enable);
  tbTop->EnableTool(wxID_ZOOM_FIT, enable);
  tbTop->EnableTool(ID_BUT_DEFAULT, enable);

  if (mode == STANDALONE) {
    wxMenuBar* mbMenu = this->GetMenuBar();
    mbMenu->Enable(wxID_CLOSE, enable);
    mbMenu->Enable(wxID_SAVE, enable);
    mbMenu->Enable(wxID_SAVEAS, enable);
    mbMenu->Enable(wxID_ZOOM_OUT, enable);
    mbMenu->Enable(wxID_ZOOM_IN, enable);
    mbMenu->Enable(wxID_ZOOM_100, enable);
    mbMenu->Enable(wxID_ZOOM_FIT, enable);
    mbMenu->Enable(ID_BUT_DEFAULT, enable);
  }

  if (!enable) {
    thumbPane->setModel(NULL);
    prePane->setModel(NULL);
    postPane->setModel(NULL);
    mixerPane->setModel(NULL);
  }
}

BEGIN_EVENT_TABLE(TintWindow, wxFrame)
//EVT_AUINOTEBOOK_ALLOW_DND(wxID_ANY, TintWindow::OnAllowNotebookDnD)
EVT_CLOSE(TintWindow::OnSystemClose)

EVT_TOOL(wxID_OPEN, TintWindow::OnButtonOpen)
EVT_TOOL(wxID_CLOSE, TintWindow::OnButtonClose)
EVT_TOOL(wxID_SAVE, TintWindow::OnButtonSave)
EVT_TOOL(wxID_SAVEAS, TintWindow::OnButtonSaveAs)
EVT_TOOL(wxID_EXIT, TintWindow::OnButtonExit)
EVT_TOOL(wxID_ZOOM_OUT, TintWindow::OnButZoomOut)
EVT_TOOL(wxID_ZOOM_IN, TintWindow::OnButZoomIn)
EVT_TOOL(wxID_ZOOM_100, TintWindow::OnButZoomNormal)
EVT_TOOL(wxID_ZOOM_FIT, TintWindow::OnButZoomFit)
EVT_TOOL(ID_BUT_DEFAULT, TintWindow::OnButtonDefault)
EVT_TOOL(wxID_ABOUT, TintWindow::OnButtonAbout)
EVT_TOOL(wxID_HELP, TintWindow::OnButtonHelp)
EVT_TOOL(ID_BUT_REGISTER, TintWindow::OnButtonRegister)
EVT_TOOL(ID_BUT_WINDOW, TintWindow::OnButtonWindow)

#ifndef __WXOSX__
EVT_MENU_OPEN(TintWindow::OnMenuOpen)

/* these events handled in OpenRecentMenu on Mac */
EVT_MENU(ID_CLEAR_RECENT, TintWindow::OnClearRecent)

/* these events handled in WindowMenu on Mac */
EVT_MENU(ID_PREPROCESSING, TintWindow::OnPreProcessPane)
EVT_MENU(ID_POSTPROCESSING, TintWindow::OnPostProcessPane)
EVT_MENU(ID_MIXER, TintWindow::OnMixerPane)
EVT_MENU(ID_THUMBS, TintWindow::OnThumbsPane)
EVT_MENU(ID_DEFAULTS, TintWindow::OnRestoreDefaults)
#endif

EVT_MENU(wxID_OPEN, TintWindow::OnButtonOpen)
EVT_MENU(wxID_CLOSE, TintWindow::OnButtonClose)
EVT_MENU(wxID_SAVE, TintWindow::OnButtonSave)
EVT_MENU(wxID_SAVEAS, TintWindow::OnButtonSaveAs)
EVT_MENU(wxID_EXIT, TintWindow::OnButtonExit)
EVT_MENU(wxID_ZOOM_OUT, TintWindow::OnButZoomOut)
EVT_MENU(wxID_ZOOM_IN, TintWindow::OnButZoomIn)
EVT_MENU(wxID_ZOOM_100, TintWindow::OnButZoomNormal)
EVT_MENU(wxID_ZOOM_FIT, TintWindow::OnButZoomFit)
EVT_MENU(ID_BUT_DEFAULT, TintWindow::OnButtonDefault)
EVT_MENU(wxID_ABOUT, TintWindow::OnButtonAbout)
EVT_MENU(wxID_HELP, TintWindow::OnButtonHelp)
EVT_MENU(ID_BUT_REGISTER, TintWindow::OnButtonRegister)

EVT_MENU(300, TintWindow::OnOpenRecent)
EVT_MENU(301, TintWindow::OnOpenRecent)
EVT_MENU(302, TintWindow::OnOpenRecent)
EVT_MENU(303, TintWindow::OnOpenRecent)
EVT_MENU(304, TintWindow::OnOpenRecent)
EVT_MENU(305, TintWindow::OnOpenRecent)
EVT_MENU(306, TintWindow::OnOpenRecent)
EVT_MENU(307, TintWindow::OnOpenRecent)

EVT_BUTTON(wxID_OK, TintWindow::OnButtonOk)
EVT_BUTTON(wxID_CANCEL, TintWindow::OnButtonCancel)
EVT_BUTTON(wxID_HELP, TintWindow::OnButtonHelp)

EVT_AUINOTEBOOK_PAGE_CLOSE(wxID_ANY, TintWindow::OnPageClose)
EVT_AUINOTEBOOK_PAGE_CHANGED(wxID_ANY, TintWindow::OnPageChanged)

END_EVENT_TABLE()
