/**
 * Copyright (C) 2007-2011 Lawrence Murray
 *
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by the Free
 * Software Foundation; either version 2 of the License, or (at your option)
 * any later version.
 * 
 * @author Lawrence Murray <lawrence@indii.org>
 * $Rev: 308 $
 * $Date: 2011-10-11 21:56:35 +0800 (Tue, 11 Oct 2011) $
 */
#ifndef INDII_TINT_GUI_TINTWINDOW_HPP
#define INDII_TINT_GUI_TINTWINDOW_HPP

#include "PreviewPane.hpp"
#include "ThumbPane.hpp"
#include "PreProcessPane.hpp"
#include "PostProcessPane.hpp"
#include "ChannelMixerPane.hpp"
#include "TintWindowObserver.hpp"
#include "TintDocument.hpp"
#include "../model/ClusterModel.hpp"
#include "../../gui/Registration.hpp"
#include "../../image/ImageResource.hpp"

#include "wx/frame.h"
#include "wx/aui/aui.h"

namespace indii {
  class OpenRecentMenu;
  class WindowMenu;

/**
 * Main window.
 */
class TintWindow : public wxFrame {
    friend class WindowMenu;
public:
  /**
   * Modes.
   */
  enum Mode {
    /**
     * Standalone mode. Includes Save and Save As buttons.
     */
    STANDALONE,
    
    /**
     * Embedded mode. Includes OK button.
     */
    EMBED
  };

  /**
   * Constructor.
   *
   * @param parent Parent.
   * @param mode Mode of operation.
   * @param obs Enclosing application, if not embedded mode.
   */
  TintWindow(wxWindow* parent, const Mode mode = EMBED,
      TintWindowObserver* obs = NULL);
  
  /**
   * Constructor.
   *
   * @param parent Parent.
   * @param res Image resource.
   * @param model Cluster model.
   * @param mode Mode of operation.
   * @param obs Enclosing application, if not embedded mode.
   */
  TintWindow(wxWindow* parent, ImageResource* res, ClusterModel* model,
      const Mode mode = EMBED, TintWindowObserver* obs = NULL);

  /**
   * Constructor.
   *
   * @param parent Parent.
   * @param path Path of image to open.
   * @param mode Mode of operation.
   * @param obs Enclosing application, if not embedded mode.
   */
  TintWindow(wxWindow* parent, const wxString& path, const Mode = STANDALONE,
      TintWindowObserver* obs = NULL);

  /**
   * Destructor.
   */
  ~TintWindow();

  /**
   * Restore default layout.
   */
  void layoutDefault();

  /**
   * Show thumbs pane.
   */
  void showThumbPane(const bool show);

  /**
   * Show preprocessing pane.
   */
  void showPrePane(const bool show);

  /**
   * Show postprocessing pane.
   */
  void showPostPane(const bool show);

  /**
   * Show mixer pane.
   */
  void showMixerPane(const bool show);

  /**
   * Is thumbs pane shown?
   */
  bool isThumbPaneShown();

  /**
   * Is preprocessing pane shown?
   */
  bool isPrePaneShown();
 
  /**
   * Is postprocessing pane shown?
   */
  bool isPostPaneShown();

  /**
   * Is mixer pane shown?
   */
  bool isMixerPaneShown();

  /**
   * Get return code.
   */
  int GetReturnCode();

  /**
   * Set return code.
   */
  void SetReturnCode(const int returnCode);

  /**
   * Close event.
   */
  void OnSystemClose(wxCloseEvent& evt);

  /**
   * Drag and drop permission event.
   */
  void OnAllowNotebookDnD(wxAuiNotebookEvent& evt);

  /**
   * Menu opened event.
   */
  void OnMenuOpen(wxMenuEvent& evt);

  /**
   * Open button event.
   */
  void OnButtonOpen(wxCommandEvent& evt);
  
  /**
   * Open recent menu event.
   */
  void OnOpenRecent(wxCommandEvent& evt);

  /**
   * Clear recent menu event.
   */
  void OnClearRecent(wxCommandEvent& evt);

  /**
   * Save button event.
   */
  void OnButtonSave(wxCommandEvent& evt);

  /**
   * Save As button event.
   */
  void OnButtonSaveAs(wxCommandEvent& evt);

  /**
   * Close button event.
   */
  void OnButtonClose(wxCommandEvent& evt);

  /**
   * Exit button event.
   */
  void OnButtonExit(wxCommandEvent& evt);
  
  /**
   * OK button event.
   */
  void OnButtonOk(wxCommandEvent& evt);

  /**
   * Cancel button event.
   */
  void OnButtonCancel(wxCommandEvent& evt);

  /*
   * Toolbar button events.
   */
  void OnButtonAbout(wxCommandEvent& evt);
  void OnButtonRegister(wxCommandEvent& evt);
  void OnButtonWindow(wxCommandEvent& evt);
  void OnButtonHelp(wxCommandEvent& evt);
  void OnButtonDefault(wxCommandEvent& evt);
  void OnButZoomOut(wxCommandEvent& evt);
  void OnButZoomIn(wxCommandEvent& evt);
  void OnButZoomNormal(wxCommandEvent& evt);
  void OnButZoomFit(wxCommandEvent& evt);
  void OnPageClose(wxAuiNotebookEvent& evt);
  void OnPageChanged(wxAuiNotebookEvent& evt);

  /*
   * Window menu events. 
   */
  void OnPreProcessPane(wxCommandEvent& evt);
  void OnPostProcessPane(wxCommandEvent& evt);
  void OnMixerPane(wxCommandEvent& evt);
  void OnThumbsPane(wxCommandEvent& evt);
  void OnRestoreDefaults(wxCommandEvent& evt);

private:
  /**
   * Enclosing application.
   */
  TintWindowObserver* obs;
  
  /**
   * Mode of operation.
   */
  Mode mode;

  /**
   * AUI manager.
   */
  wxAuiManager* manager;
  
  /**
   * Notebook, for standalone mode.
   */
  wxAuiNotebook* notebook;

  /**
   * Document, for embedded mode.
   */
  TintDocument* doc;

  /**
   * Return code.
   */
  int returnCode;
  
  /*
   * Menus.
   */
  OpenRecentMenu* mnRecent;
  WindowMenu* mnWindow;
  WindowMenu* mnPopupWindow;
  wxMenu* mnFile;

  /*
   * Panes.
   */
  ThumbPane* thumbPane;
  PreProcessPane* prePane;
  PostProcessPane* postPane;
  ChannelMixerPane* mixerPane;
  
  /**
   * Initialise.
   *
   * @param mode Mode of operation.
   * @param doc Document to have open initially.
   */
  void init(const Mode mode, TintDocument* doc = NULL);

  /**
   * Enable or disable menu and toolbar items as appropriate.
   */
  void able(const bool enable);

  /*
   * Control ids.
   */
  enum ID {
    ID_BUT_DEFAULT = 1, // can't have menu item id of 0 on Mac
    ID_BUT_REGISTER,
    ID_BUT_WINDOW,

    /* these must match those given in OpenRecentMenu */
    ID_CLEAR_RECENT = 764,
    ID_OPEN_RECENT = 765,

    /* these must match those given in WindowMenu */
    ID_PREPROCESSING = 1432,
    ID_POSTPROCESSING,
    ID_MIXER,
    ID_THUMBS,
    ID_DEFAULTS,\
  };
  
  /**
   * Extra height for floating panels.
   */
  #ifdef __WXMAC__
  static const int EXTRA_PANE_WIDTH = 0;
  static const int EXTRA_PANE_HEIGHT = 16;
  #else
  static const int EXTRA_PANE_WIDTH = 0;
  static const int EXTRA_PANE_HEIGHT = 0;
  #endif
  static const int MIN_THUMB_PANE_WIDTH = 387;

  DECLARE_EVENT_TABLE()
};
}

inline void indii::TintWindow::showThumbPane(const bool show) {
  manager->GetPane(thumbPane).Show(show);
  manager->Update();
}

inline void indii::TintWindow::showPrePane(const bool show) {
  manager->GetPane(prePane).Show(show);
  manager->Update();
}

inline void indii::TintWindow::showPostPane(const bool show) {
  manager->GetPane(postPane).Show(show);
  manager->Update();
}

inline void indii::TintWindow::showMixerPane(const bool show) {
  manager->GetPane(mixerPane).Show(show);
  manager->Update();
}

inline bool indii::TintWindow::isThumbPaneShown() {
  return manager->GetPane(thumbPane).IsShown();
}

inline bool indii::TintWindow::isPrePaneShown() {
  return manager->GetPane(prePane).IsShown();
}
 
inline bool indii::TintWindow::isPostPaneShown() {
  return manager->GetPane(postPane).IsShown();
}

inline bool indii::TintWindow::isMixerPaneShown() {
  return manager->GetPane(mixerPane).IsShown();
}

#endif
